# Copyright (C) 2021 Fraunhofer Institut IOSB, Fraunhoferstr. 1, D 76131
# Karlsruhe, Germany.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import json

from frost_sta_client.dao.actuator import ActuatorDao

from . import entity
from . import tasking_capability

from .ext import entity_list
from .ext import entity_type
from frost_sta_client import utils


class Actuator(entity.Entity):

    def __init__(self,
                 name='',
                 description='',
                 encoding_type='',
                 tasking_capabilities=None,
                 metadata='',
                 properties=None,
                 **kwargs):
        super().__init__(**kwargs)
        if properties is None:
            properties = {}
        self.name = name
        self.description = description
        self.encoding_type = encoding_type
        self.tasking_capabilities = tasking_capabilities
        self.metadata = metadata
        self.properties = properties

    def __new__(cls, *args, **kwargs):
        new_actuator = super().__new__(cls)
        attributes = {'_id': None, '_name': '', '_description': '', '_properties': {}, '_encoding_type': '',
                      '_metadata': '', '_self_link': '', '_service': None, '_tasking_capabilities': None}
        for key, value in attributes.items():
            new_actuator.__dict__[key] = value
        return new_actuator

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if not isinstance(value, str):
            raise ValueError('name should be of type str!')
        self._name = value

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value):
        if not isinstance(value, str):
            raise ValueError('description should be of type str!')
        self._description = value

    @property
    def encoding_type(self):
        return self._encoding_type

    @encoding_type.setter
    def encoding_type(self, value):
        if not isinstance(value, str):
            raise ValueError('encodingtype should be of type str!')
        self._encoding_type = value

    @property
    def metadata(self):
        return self._metadata

    @metadata.setter
    def metadata(self, value):
        try:
            json.dumps(value)
        except TypeError:
            raise TypeError('result should be json serializable')
        self._metadata = value
        if self._metadata is None:
            raise Warning('metadata is a mandatory property')

    @property
    def properties(self):
        return self._properties

    @properties.setter
    def properties(self, value):
        if not isinstance(value, dict):
            raise ValueError('properties should be of type dict!')
        self._properties = value

    @property
    def tasking_capabilities(self):
        return self._tasking_capabilities

    @tasking_capabilities.setter
    def tasking_capabilities(self, values):
        if values is None:
            self._tasking_capabilities = None
            return
        if isinstance(values, list) and all(isinstance(tc, tasking_capability.TaskingCapability) for tc in values):
            entity_class = entity_type.EntityTypes['TaskingCapability']['class']
            self._tasking_capabilities = entity_list.EntityList(entity_class=entity_class, entities=values)
            return
        if not isinstance(values, entity_list.EntityList) or \
                any((not isinstance(tc, tasking_capability.TaskingCapability)) for tc in values.entities):
            raise ValueError('Tasking capabilities should be a list of TaskingCapabilities')
        self._tasking_capabilities = values

    def ensure_service_on_children(self, service):
        if self.tasking_capabilities is not None:
            self.tasking_capabilities.set_service(service)

    def __eq__(self, other):
        if other is None:
            return False
        if not isinstance(other, type(self)):
            return False
        if id(self) == id(other):
            return True
        if self.name != other.name:
            return False
        if self.description != other.description:
            return False
        if self.encoding_type != other.encoding_type:
            return False
        if self.metadata != other.metadata:
            return False
        if self.properties != other.properties:
            return False
        return True

    def __ne__(self, other):
        return not self == other

    def __getstate__(self):
        data = super().__getstate__()
        if data.get('@iot.id', None) is not None:
            return data
        if self.name is not None and self.name != '':
            data['name'] = self.name
        if self.description is not None and self.description != '':
            data['description'] = self.description
        if self.encoding_type is not None:
            data['encodingType'] = self.encoding_type
        if self.metadata is not None:
            data['metadata'] = self.metadata
        if self.properties is not None and self.properties != {}:
            data['properties'] = self.properties
        if self.tasking_capabilities is not None and len(self.tasking_capabilities.entities) > 0:
            data['taskingCapabilities'] = self.tasking_capabilities.__getstate__()
        return data

    def __setstate__(self, state):
        super().__setstate__(state)
        self.name = state.get("name", None)
        self.description = state.get("description", None)
        self.encoding_type = state.get("encodingType", "")
        self.metadata = state.get("metadata", "")
        self.properties = state.get("properties", None)
        if state.get("TaskingCapabilities", None) is not None and isinstance(state["TaskingCapabilities"], list):
            entity_class = entity_type.EntityTypes['TaskingCapability']['class']
            self.tasking_capabilities = utils.transform_json_to_entity_list(state['TaskingCapabilities'], entity_class)
            self.tasking_capabilities.next_link = state.get("TaskingCapabilities@iot.nextLink", None)

    def get_dao(self, service):
        return ActuatorDao(service)



from frost_sta_client.model import actuator
from frost_sta_client.model import datastream
from frost_sta_client.model import entity
from frost_sta_client.model import feature_of_interest
from frost_sta_client.model import historical_location
from frost_sta_client.model import location
from frost_sta_client.model import multi_datastream
from frost_sta_client.model import observation
from frost_sta_client.model import observedproperty
from frost_sta_client.model import sensor
from frost_sta_client.model import task
from frost_sta_client.model import tasking_capability
from frost_sta_client.model import thing



# Copyright (C) 2021 Fraunhofer Institut IOSB, Fraunhoferstr. 1, D 76131
# Karlsruhe, Germany.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from abc import ABC
from frost_sta_client.service.sensorthingsservice import SensorThingsService


class Entity(ABC):
    """
    An abstract representation of an entity.
    """
    def __init__(self,
                 id=None,
                 self_link='',
                 service=None):
        self.id = id
        self.self_link = self_link
        self.service = service

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        if value is None:
            self._id = None
            return
        if isinstance(value, int) or isinstance(value, str):
            self._id = value
            return
        raise ValueError('id of entity should be of type int!')

    @property
    def self_link(self):
        return self._self_link

    @self_link.setter
    def self_link(self, value):
        if not isinstance(value, str):
            raise ValueError('self_link should be of type str!')
        self._self_link = value

    @property
    def service(self):
        return self._service

    @service.setter
    def service(self, value):
        if value is None or isinstance(value, SensorThingsService):
            self._service = value
            return
        raise ValueError('service should be of type SensorThingsService')

    def set_service(self, service):
        if self.service != service:
            self.service = service
            self.ensure_service_on_children(service)

    @property
    def IOT_COUNT(self):
        return 'iot.count'

    @property
    def AT_IOT_COUNT(self):
        return '@iot.count'

    @property
    def IOT_NAVIGATION_LINK(self):
        return 'iot.navigationLink'

    @property
    def AT_IOT_NAVIGATION_LINK(self):
        return '@iot.navigationLink'

    @property
    def IOT_NEXT_LINK(self):
        return 'iot.nextLink'

    @property
    def AT_IOT_NEXT_LINK(self):
        return '@iot.nextLink'

    @property
    def IOT_SELF_LINK(self):
        return 'iot.selfLink'

    @property
    def AT_IOT_SELF_LINK(self):
        return '@iot.selfLink'

    def __eq__(self, other):
        if other is None:
            return False
        if not isinstance(other, type(self)):
            return False
        if id(self) == id(other):
            return True
        if self.id != other.id:
            return False
        if self.self_link != other.self_link:
            return False
        return True

    def __ne__(self, other):
        return not self == other

    def __getstate__(self):
        data = {}
        if self.id is not None and self.id != '':
            data['@iot.id'] = self.id
        return data

    def __setstate__(self, state):
        self.id = state.get('@iot.id', None)
        self.self_link = state.get('@iot.selfLink', '')



import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
info = {}
with open(os.path.join(here, 'frost_sta_client', '__version__.py')) as f:
    exec(f.read(), info)

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name=info['__title__'],
    version=info['__version__'],
    description=info['__description__'],
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=info['__author__'],
    author_email=info['__contact__'],
    license=info['__license__'],
    url=info['__url__'],
    packages=find_packages(),
    install_requires=['demjson3>=3.0.5', 'furl>=2.1.3', 'geojson>=2.5.0', 'jsonpickle>=2.0.0', 'requests>=2.26.0',
                      'jsonpatch', 'python-dateutil'],
    keywords=['sta', 'ogc', 'frost', 'sensorthingsapi', 'IoT']
)



# Copyright (C) 2021 Fraunhofer Institut IOSB, Fraunhoferstr. 1, D 76131
# Karlsruhe, Germany.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json

import frost_sta_client.model
from . import entity
from . import multi_datastream
from . import datastream
from . import feature_of_interest

from frost_sta_client.dao.observation import ObservationDao

from frost_sta_client import utils


class Observation(entity.Entity):

    def __init__(self,
                 phenomenon_time=None,
                 result=None,
                 result_time=None,
                 result_quality=None,
                 valid_time=None,
                 parameters=None,
                 datastream=None,
                 multi_datastream=None,
                 feature_of_interest=None,
                 **kwargs):
        super().__init__(**kwargs)
        if parameters is None:
            parameters = {}
        self.phenomenon_time = phenomenon_time
        self.result = result
        self.result_time = result_time
        self.result_quality = result_quality
        self.valid_time = valid_time
        self.parameters = parameters
        self.datastream = datastream
        self.multi_datastream = multi_datastream
        self.feature_of_interest = feature_of_interest

    def __new__(cls, *args, **kwargs):
        new_observation = super().__new__(cls)
        attributes = {'_id': None, '_phenomenon_time': None, '_result': None, '_result_time': None,
                      '_result_quality': None, '_valid_time': None, '_parameters': {}, '_datastream': None,
                      '_multi_datastream': None, '_feature_of_interest': None, '_self_link': '', '_service': None}
        for key, value in attributes.items():
            new_observation.__dict__[key] = value
        return new_observation

    @property
    def phenomenon_time(self):
        return self._phenomenon_time

    @phenomenon_time.setter
    def phenomenon_time(self, value):
        self._phenomenon_time = utils.check_datetime(value, 'phenomenon_time')

    @property
    def result(self):
        return self._result

    @result.setter
    def result(self, value):
        if value is None:
            self._result = None
            return
        try:
            json.dumps(value)
        except TypeError:
            raise TypeError('result should be json serializable')
        self._result = value

    @property
    def result_time(self):
        return self._result_time

    @result_time.setter
    def result_time(self, value):
        self._result_time = utils.check_datetime(value, 'result_time')

    @property
    def result_quality(self):
        return self._result_quality

    @result_quality.setter
    def result_quality(self, value):
        if value is None:
            self._result_quality = None
            return
        try:
            json.dumps(value)
        except TypeError:
            raise TypeError('result_quality should be json serializable')
        self._result_quality = value

    @property
    def valid_time(self):
        return self._valid_time

    @valid_time.setter
    def valid_time(self, value):
        self._valid_time = utils.check_datetime(value, 'valid_time')

    @property
    def parameters(self):
        return self._parameters

    @parameters.setter
    def parameters(self, values):
        if values is None:
            self._parameters = None
            return
        if not isinstance(values, dict):
            raise ValueError('parameters should be of type dict!')
        self._parameters = values

    @property
    def feature_of_interest(self):
        return self._feature_of_interest

    @feature_of_interest.setter
    def feature_of_interest(self, value):
        if value is None:
            self._feature_of_interest = None
            return
        if not isinstance(value, feature_of_interest.FeatureOfInterest):
            raise ValueError('feature_of_interest should be of type FeatureOfInterest!')
        self._feature_of_interest = value

    @property
    def datastream(self):
        return self._datastream

    @datastream.setter
    def datastream(self, value):
        if value is None:
            self._datastream = None
            return
        if not isinstance(value, datastream.Datastream):
            raise ValueError('datastream should be of type Datastream!')
        self._datastream = value

    @property
    def multi_datastream(self):
        return self._multi_datastream

    @multi_datastream.setter
    def multi_datastream(self, value):
        if value is None:
            self._multi_datastream = None
            return
        if isinstance(value, multi_datastream.MultiDatastream):
            self._multi_datastream = value
            return
        raise ValueError('multi_datastream should be of type MultiDatastream!')

    def ensure_service_on_children(self, service):
        if self.datastream is not None:
            self.datastream.set_service(service)
        if self.multi_datastream is not None:
            self.multi_datastream.set_service(service)
        if self.feature_of_interest is not None:
            self.feature_of_interest.set_service(service)

    def __eq__(self, other):
        if other is None:
            return False
        if not isinstance(other, type(self)):
            return False
        if id(self) == id(other):
            return True
        if self.result != other.result:
            return False
        if self.phenomenon_time != other.phenomenon_time:
            return False
        if self.result_time != other.result_time:
            return False
        if self.valid_time != other.valid_time:
            return False
        if self.parameters != other.parameters:
            return False
        if self.result_quality != other.result_quality:
            return False
        return True

    def __ne__(self, other):
        return not self == other

    def __getstate__(self):
        data = super().__getstate__()
        if data.get('@iot.id', None) is not None:
            return data
        if self.parameters is not None and self.parameters != {}:
            data['parameters'] = self.parameters
        if self.result is not None:
            data['result'] = self.result
        if self.result_quality is not None:
            data['resultQuality'] = self.result_quality
        if self.phenomenon_time is not None:
            data['phenomenonTime'] = utils.parse_datetime(self.phenomenon_time)
        if self.result_time is not None:
            data['resultTime'] = utils.parse_datetime(self.result_time)
        if self.valid_time is not None:
            data['validTime'] = utils.parse_datetime(self.valid_time)
        if self.datastream is not None:
            data['Datastream'] = self.datastream.__getstate__()
        if self.multi_datastream is not None:
            data['MultiDatastream'] = self.multi_datastream.__getstate__()
        if self.feature_of_interest is not None:
            data['FeatureOfInterest'] = self.feature_of_interest.__getstate__()
        return data

    def __setstate__(self, state):
        super().__setstate__(state)
        self.parameters = state.get("parameters", {})
        self.result = state.get("result", None)
        self.result_quality = state.get("resultQuality", None)
        self.phenomenon_time = state.get("phenomenonTime", None)
        self.result_time = state.get("resultTime", None)
        self.valid_time = state.get("validTime", None)
        if state.get('Datastream', None) is not None:
            self.datastream = frost_sta_client.model.datastream.Datastream()
            self.datastream.__setstate__(state['Datastream'])
        if state.get('MultiDatastream', None) is not None:
            self.multi_datastream = frost_sta_client.model.multi_datastream.MultiDatastream()
            self.multi_datastream.__setstate__(state['MultiDatastream'])
        if state.get("FeatureOfInterest", None) is not None:
            self.feature_of_interest = frost_sta_client.model.feature_of_interest.FeatureOfInterest()
            self.feature_of_interest.__setstate__(state['FeatureOfInterest'])

    def get_dao(self, service):
        return ObservationDao(service)



# Copyright (C) 2021 Fraunhofer Institut IOSB, Fraunhoferstr. 1, D 76131
# Karlsruhe, Germany.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import jsonpickle
import demjson3
import datetime
from dateutil.parser import isoparse
import geojson
import logging
import frost_sta_client.model.ext.entity_list


def extract_value(location):
    return location[location.find('(')+1: location.find(')')]


def transform_entity_to_json_dict(entity):
    json_str = jsonpickle.encode(entity, unpicklable=False)
    return jsonpickle.decode(json_str)


def transform_json_to_entity(json_response, entity_class):
    decodable_str = '{\'py/object\': \'' + entity_class + '\', \'py/state\': ' \
                    + jsonpickle.encode(json_response, unpicklable=False) + '}'
    return jsonpickle.decode(decodable_str, backend=demjson3)


def transform_json_to_entity_list(json_response, entity_class):
    entity_list = frost_sta_client.model.ext.entity_list.EntityList(entity_class)
    result_list = []
    if isinstance(json_response, dict):
        try:
            response_list = json_response['value']
            entity_list.next_link = json_response.get("@iot.nextLink", None)
        except AttributeError as e:
            raise e
    else:
        response_list = json_response
    for item in response_list:
        result_list.append(transform_json_to_entity(item, entity_list.entity_class))
    entity_list.entities = result_list
    return entity_list


def check_datetime(value, time_entity):
    try:
        parse_datetime(value)
    except ValueError as e:
        logging.error(f"error during {time_entity} check")
        raise e
    return value


def parse_datetime(value) -> str:
    if value is None:
        return value
    if isinstance(value, str):
        if '/' in value:
            try:
                times = value.split('/')
                if len(times) != 2:
                    raise ValueError("If the time interval is provided as a string,"
                                     " it should be in isoformat")
                result = [isoparse(times[0]),
                          isoparse(times[1])]
            except ValueError:
                raise ValueError("If the time entity interval is provided as a string,"
                                 " it should be in isoformat")
            result = result[0].isoformat() + '/' + result[1].isoformat()
            return result
        else:
            try:
                result = isoparse(value)
            except ValueError:
                raise ValueError("If the phenomenon time is provided as string, it should be in isoformat")
            result = result.isoformat()
            return result
    if isinstance(value, datetime.datetime):
        return value.isoformat()
    if isinstance(value, list) and all(isinstance(v, datetime.datetime) for v in value):
        return value[0].isoformat() + value[1].isoformat()
    else:
        raise ValueError('time entities should consist of one or two datetimes')



def process_area(value):
    if not isinstance(value, dict):
        raise ValueError("geojsons can only be handled as dictionaries!")
    if value.get("type", None) is None or value.get("coordinates", None) is None:
        raise ValueError("Both type and coordinates need to be specified in the dictionary")
    if value["type"] == "Point":
        return geojson.geometry.Point(value["coordinates"])
    if value["type"] == "Polygon":
        return geojson.geometry.Polygon(value["coordinates"])
    if value["type"] == "Geometry":
        return geojson.geometry.Geometry(value["coordinates"])
    raise ValueError("can only handle geojson of type Point, Polygon or Geometry")
